#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''Extend PyTado interface to provide more information easily.'''

from PyTado.interface import Tado


class TadoExtended(Tado):
    '''Extends the PyTado interface to provide additional methods.'''

    def getHeatingState(self, zone):
        '''Gets set temperature (centigrade) and heating power (%) for Zone zone.'''

        data = self.getState(zone)
        return {'temperature': data['setting']['temperature']['celsius'],
                'power': data['activityDataPoints']['heatingPower']['percentage']}
    
    def getOpenWindow(self, zone):
        '''Gets open window status for Zone zone.'''

        if self.getState(zone)['openWindow'] is None:
            return False
        else:
            return True
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''Contains datalogger classes for PyTado.'''

import concurrent.futures
import getpass
import json
import logging
import multiprocessing
import pathlib
import sched
import time
from urllib.error import HTTPError, URLError

import keyring
import numpy as np

from .datastores import DataStore, MPDataStore
from .extended_interface import TadoExtended
from .utils import nexttick, loglistener, setuplogger, setupoutdir


_LOGGER = logging.getLogger(__name__)


class TadoLogger:
    '''Periodically requests TaDo data and passes to dataframe.'''
    CREDPATH = pathlib.Path.home() / '.tado_credentials'

    def __init__(self, outdir=None, update_period=30, last_day='sun', multiproc=True, loglevel=logging.WARNING):
        '''Constructs TaDo logger to update at update_periods.

        Args:
            outdir (path like, optional): Output direct to store csvs and logs.
                Defaults to None which places files in <home>/Documents/TadoLogs/.
            update_period (int, optional): Time in seconds between update.
                Defaults to 30.
            last_day (str, optional): Three letter weekday string indicating
                last day of week stored in csv files. Defaults to 'sun'.
            multiproc (bool, optional): Run datastore in another process. 
                Defaults to True.
            loglevel (logging level, optional): Level to log at. Defaults to
                logging.WARNING.
        '''
        self.update_period = update_period
        self.lastday = last_day
        self._multiproc = multiproc
        self._loglevel = loglevel
        self._outdir = setupoutdir(outdir)
        self.event = None  # Will hold next update event
        self.scheduler = sched.scheduler(time.time, time.sleep)
        self._logqueue = multiprocessing.Queue(-1)
        self._stopev = multiprocessing.Event()
        self._loglistener = multiprocessing.Process(
            target=loglistener,
            args=(self._outdir, self._logqueue, self._stopev, self._loglevel),
            )
        self._loglistener.start()
        setuplogger(_LOGGER, self._logqueue, self._loglevel)
        _LOGGER.info('---STARTED TADOLOGGER---')
        self.variables = {
            'Weather': ('Outside Temp (°C)', 'Solar Int. (%)', 'Weather'),
            'Zones': (
                'Temp (°C)', 'R.H. (%)',
                'Set Temp (°C)', 'Heating Power (%)',
                'Open Window'
                ),
            }
        _LOGGER.info('Logging into TaDo server')
        self.login()
        self.zones = sorted(self.home.getZones(), key=lambda z: z["id"])
        _LOGGER.debug('Found %s zones in home', len(self.zones))

        if self._multiproc:
            self._q = multiprocessing.JoinableQueue()
            self.pdstore = MPDataStore(
                self.variables,
                self.zones,
                self._outdir,
                self.update_period,
                self.lastday,
                self._logqueue,
                self._loglevel,
                self._q,
                self._stopev,
                )
        else:
            self.pdstore = DataStore(
                self.variables,
                self.zones,
                self._outdir,
                self.update_period,
                self.lastday,
                self._logqueue,
                self._loglevel,
                )
    
    def __enter__(self):
        _LOGGER.debug('PyTadoLog context manager entered')
        return self
    
    def __exit__(self, exc_type, exc_value, exc_traceback):
        _LOGGER.debug('PyTadoLog context manager exiting')
        if exc_type is not None:
            _LOGGER.exception(
                'Error encountered during operation',
                exc_info=(exc_type, exc_value, exc_traceback),
                )
        self.close()

    def login(self):
        '''Log in to TaDo mobile server.

        Check if credentials are stored using default OS credentials manager
        and prompts for new credentials if required.
        '''
        if not self.CREDPATH.exists():
            _LOGGER.error(
                'Credentials file cannot be found - Enter Tado credentials',
                )
            self.setcredentials()
        
        loggedin = False
        while not loggedin:
            try:
                with open(self.CREDPATH, mode='r') as f:
                    credentials = json.load(f)
            except json.decoder.JSONDecodeError:
                _LOGGER.error(
                    'Credentials file is corrupted - reenter Tado credentials',
                    )
                self.setcredentials()
            else:
                try:
                    self.home = TadoExtended(
                        credentials['username'],
                        keyring.get_password(
                            credentials['service'], 
                            credentials['username']
                            )
                        )
                except HTTPError:
                    _LOGGER.error(
                        'Tado username or password is incorrect - reenter credentials',
                        )
                    self.setcredentials()
                else:
                    loggedin = True
                    _LOGGER.debug('Logged into TaDo server')

    def setcredentials(self, service='tado'):
        '''Prompt user to enter TaDo login credentials.

        Args:
            service (str, optional): Name of TaDo service in default OS 
                credentials manager. Defaults to 'tado'.
        
        Stores the service name and username in self.CREDPATH.
        '''
        _LOGGER.debug('Requesting TaDo username from user')
        username = str(input('Enter Tado username: '))
        with open(self.CREDPATH, 'w') as f:
            _LOGGER.debug(
                'Saving TaDo credentials details to %s', self.CREDPATH
                )
            json.dump({'service': service, 'username': username}, f)
        keyring.set_password(service, username, getpass.getpass())
        _LOGGER.debug('TaDo credentials stored in OS credentials manager')
    
    def update(self, t):
        '''Get new data and append to dataframe then schedule next update.

        Args:
            t (datetime.datetime): Scheduled time of current update.
        '''
        if self._multiproc:
            self._q.put((t, self.getnewdata()))
        else:
            self.pdstore.update(t, self.getnewdata())

        # Schedule the next update
        start_time = nexttick(self.update_period)
        _LOGGER.debug('Update scheduled for %s', start_time)
        print(f'Next update scheduled for {start_time}', end='\r', flush=True)
        self.event = self.scheduler.enterabs(
            time.mktime(start_time.timetuple()),
            1,
            self.update,
            argument = (start_time, ),
            )
    
    def getnewdata(self):
        '''Get latest data from TaDo server and add to dataframe.
        
        Uses multithreaded pool to get data simultaneously.
        '''
        with concurrent.futures.ThreadPoolExecutor(
                max_workers = len(self.zones) + 1
                ) as executor:
            # Load the operations then wait for them all to complete
            futures = [executor.submit(self.getweatherdata)]
            futures.extend(
                [executor.submit(self.getzonedata, zone) for zone in self.zones]
                )
            futures, _ = concurrent.futures.wait(futures, timeout=30)
            return [future.result() for future in futures]
    
    def getzonedata(self, zone):
        '''Get data on requested zone from TaDo server.

        Args:
            zone (int): ID for TaDo Zone.

        Returns:
            tuple: Tuple containing Zone data.

        Uses try/except clauses to handle server hiccups.
        '''
        try:
            climate = self.home.getClimate(zone['id'])
        except URLError:
            # If the server doesn't respond continue running
            _LOGGER.warning('Could not get climate for %s', zone["name"])
            climate = {'temperature': np.nan, 'humidity': np.nan}
        try:
            state = self.home.getHeatingState(zone['id'])
        except URLError:
            # If the server doesn't respond continue running
            _LOGGER.warning('Could not get heating state for %s', zone["name"])
            state = {'temperature': np.nan, 'power': np.nan}
        try:
            window = self.home.getOpenWindow(zone['id'])
        except URLError:
            # If the server doesn't respond continue running
            _LOGGER.warning('Could not get window state for %s', zone["name"])
            window = np.nan
        
        _LOGGER.debug('Got all data for %s', zone["name"])
        out = (
            zone['name'], (
                climate['temperature'],
                climate['humidity'],
                state['temperature'],
                state['power'],
                window,
                ),
            )
        return ('Zones', out)
    
    def getweatherdata(self):
        '''Get weather data from TaDo server.

        Returns:
            tuple: TaDo weather data.
        
        Uses try/except clauses to handle server hiccups.
        '''
        try:
            data = self.home.getWeather()
            dataout = (
                data['outsideTemperature']['celsius'],
                data['solarIntensity']['percentage'],
                data['weatherState']['value'],
                )
        except URLError:
            # If the server doesn't respond continue running
            _LOGGER.debug('Could not get weather data')
            dataout = (np.nan, np.nan, np.nan)
        out = ('Weather', dataout)
        return ('Weather', out)

    def start(self):
        '''Start running the data logger.

        Sets up the dataframe and schedules the first update.
        '''
        start_time = nexttick(self.update_period)
        self.event = self.scheduler.enterabs(
            time.mktime(start_time.timetuple()),
            1,
            self.update,
            argument = (start_time, ),
            )
        _LOGGER.info('First data point scheduled for %s', start_time)
        if self._multiproc:
            self.pdstore.start()
        self.scheduler.run()

    def close(self):
        '''Clean up any scheduled events.'''
        if self.scheduler is not None:
            if not self.scheduler.empty():
                _LOGGER.debug('Cancelling scheduled event')
                self.scheduler.cancel(self.event)
        
        if self._multiproc:
            self._stopev.set()
            _LOGGER.debug('Set stop event')
            self._q.join()
            _LOGGER.debug('Datastore process stopped')


if __name__ == '__main__':
    import datetime as dt
    
    with TadoLogger() as tl:
        tl.start()
    print(f'Logging ended at {dt.datetime.now()}')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''Contains datastore classes for PyTado.'''

import datetime as dt
import logging
import multiprocessing
import pathlib
import re

import numpy as np
import pandas as pd

from .utils import getnextday, nexttick, setuplogger


_LOGGER = logging.getLogger(__name__)


class DataStore:
    '''Class to manage pandas dataframe.

    CSV files are stored in /usr/Documents/TaDoLogs.
    '''
    DATEPATTERN = re.compile(r'\d{4}-\d{2}-\d{2}')

    def __init__(self, variables, zones, outdir, delta, lastday, logqueue, loglevel=logging.WARNING):
        '''Initialise datastore. Sets up columns based on vars and zones

        Args:
            variables (list of dicts): List of variables to use for columns.
            zones (List of dicts): List of zones in TaDo home.
            outdir (pathlib.Path): Directory to hold CSV files.
            delta (int): Time in seconds between rows in dataframe.
            last_day (str, optional): Three letter weekday string indicating 
                last day of week stored in csv files. Defaults to 'sun'.
            logqueue(multiprocessing.queue): Queue to hold logging.logger logs.
            loglevel (logging level, optional): Level to log at. Defaults to
                logging.WARNING.
        '''
        self.variables = variables
        self.zones = zones
        self.outdir = outdir
        self.delta = delta
        self.lastday = lastday
        self.csvpath = None  # Will hold path to current csv file
        self.lastindex = None  # Will hold last index in dataframe
        self.df = None  # Will hold dataframe
        setuplogger(_LOGGER, logqueue, loglevel)

        _LOGGER.debug('Creating output directory at %s', self.outdir)
        self.outdir.mkdir(exist_ok=True)
        self.setupdf()
    
    def setupdf(self):
        '''Sets up the dataframe object to hold TaDo data.

        Looks for existing suitable csv files and loads them if present.
        '''
        self.csvpath_setter()

        weather_cols = pd.MultiIndex.from_product(
            (['Weather'], self.variables['Weather']),
            names=['Zone', 'Variable'],
            )
        zone_cols = pd.MultiIndex.from_product(
            ([zone['name'] for zone in self.zones], self.variables['Zones']), 
            names=['Zone', 'Variable'],
            )
        columns = weather_cols.append(zone_cols)

        # Check if the supplied csv path is a file and in the correct format
        df_loaded = False
        if self.csvpath.exists():
            _LOGGER.debug('Found file at %s', self.csvpath)
            self.df = pd.read_csv(
                self.csvpath, 
                header = [0, 1], 
                index_col = 0, 
                parse_dates = True,
                )
            _LOGGER.debug('Loaded csv file into dataframe')
            if self.df.columns.equals(columns):
                _LOGGER.debug('Columns in csv file match expected format')
                extended_idx = pd.date_range(
                    start = self.df.index[0], 
                    end = self.lastindex,
                    freq=f'{self.delta}S',
                    )
                self.df = self.df.reindex(extended_idx)
                _LOGGER.debug('Extended dataframe index to %s', self.lastindex)
                # Add an empty row to indicate a gap at the next tick
                self.addemptyrow()
                df_loaded = True
            else:
                # Rename the old file so it is not overwritten and create a new
                # file with all required columns
                _LOGGER.warning('csv file does not match expected format')
                self.csvpath.rename(
                    self.csvpath.parent / f'{self.csvpath.stem}_old.csv'
                    )
                _LOGGER.info('Old csv file renamed to %s_old.csv', self.csvpath.stem)

        if not df_loaded:
            _LOGGER.debug('Creating new dataframe')
            index = pd.date_range(
                start = nexttick(self.delta),
                end = self.lastindex,
                freq=f'{self.delta}S',
                )
            self.df = pd.DataFrame(index=index, columns=columns)
            self.df.to_csv(self.csvpath)
        
    def update(self, t, data):
        '''Add a new entry to dataframe at row with index t.

        Args:
            t (datetime.datetime): Row index for new data.
            data (list of dicts): List of data dictionaries.
        '''
        if t > self.lastindex:
            self.setupdf()
        for d in data:
            key, result = d
            for i, var in enumerate(self.variables[key]):
                self.df.loc[t, (result[0], var)] = result[1][i]
        self.save()
        _LOGGER.debug('Saved to csv')
    
    def save(self, dropna=True):
        '''Save dataframe to csv file.

        Args:
            dropna (bool, optional): Drop NaN rows on save. Defaults to True.
        '''
        if dropna:
            self.df.dropna(how='all').to_csv(self.csvpath)
            _LOGGER.debug('Dropped NA rows from dataframe and saved to csv')
        else:
            self.df.to_csv(self.csvpath)
            _LOGGER.debug('Saved to csv')
    
    def csvpath_setter(self):
        '''Set path to csv datastore.

        Loooks for existing datastores and uses the latest one if available.
        If there are no suitable ones it will create a path to a new one.
        '''
        today = dt.date.today()
        viable_files = []
        _LOGGER.debug('Searching for viable csv files in %s', self.outdir)
        for path in self.outdir.glob('HomeData(*_to_*).csv'):
            start, end = [
                dt.datetime.strptime(date, '%Y-%m-%d').date() 
                for date in self.DATEPATTERN.findall(path.stem)
                ]
            if start <= today <= end:
                _LOGGER.debug('Found matching file with name %s', path.stem)
                viable_files += [(path, start, end)]
        if not viable_files:
            # Create a new file if there is no existing file
            _LOGGER.debug('No viable files found')
            self.lastindex = dt.datetime.combine(
                getnextday(self.lastday), 
                dt.time(23, 59, 59)
                )
            self.csvpath = self.outdir / f'HomeData({today}_to_{self.lastindex.date()}).csv'
        else:
            _LOGGER.debug('Found %s viable files', len(viable_files))
            if len(viable_files) > 1:
                viable_files = sorted(
                    viable_files,
                    reverse=True,
                    key=lambda tup: (tup[2], tup[1]),
                    )[0][0]
            self.lastindex = dt.datetime.combine(
                viable_files[0][2], 
                dt.time(23, 59, 59)
                )
            self.csvpath = viable_files[0][0]
        _LOGGER.debug('Using csv file with name %s', self.csvpath)
    
    def addemptyrow(self):
        '''Insert a row of inifinities to indicate a break in data logging.
        '''
        # Create row at previous delta time
        start_time = nexttick(self.delta) - dt.timedelta(seconds=self.delta)
        if start_time > self.lastindex:
            self.setupdf()
        elif start_time > self.df.index[0]:
            self.df.loc[start_time] = np.inf
            self.save()
            _LOGGER.debug('Added inf row to dataframe at %s', start_time)


class MPDataStore(multiprocessing.Process, DataStore):
    '''Multiprocessing version of datastore.
    '''
    def __init__(
            self,
            variables,
            zones,
            outdir,
            delta,
            lastday,
            logqueue,
            loglevel=logging.WARNING,
            q = None,
            stopev = None,
            ):
        '''Initialise multiprocessing datastore.

        Args:
            variables (list of dicts): List of variables to use for columns.
            zones (List of dicts): List of zones in TaDo home.
            outdir (pathlib.Path): Directory to hold CSV files.
            delta (int): Time in seconds between rows in dataframe.
            last_day (str, optional): Three letter weekday string indicating
                last day of week stored in csv files. Defaults to 'sun'.
            logqueue(Queue): Queue to hold logging.logger logs.
            loglevel (logging level, optional): Level to log at. Defaults to
                logging.WARNING.
            q (JoinableQueue, optional): Queue which will hold data. Defaults
                to None.
            stopev (multiprocessing.Event, optional): Event which will signal 
                to stop processing. Defaults to None.
        '''
        multiprocessing.Process.__init__(self)
        self._variables = variables
        self._zones = zones
        self._outdir = outdir
        self._delta = delta
        self._lastday = lastday
        self._logqueue = logqueue
        self._loglevel = loglevel
        self._q = q
        self._stopev = stopev
        
    def run(self):
        '''Wait for data in queue and update datastore when available.
        '''
        DataStore.__init__(self, self._variables, self._zones, self._outdir, self._delta, self._lastday, self._logqueue, self._loglevel)
        while not self._stopev.is_set():
            self.update(*self._q.get())
            self._q.task_done()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''Main module for running TadoLogger from commandline.'''
import argparse
import datetime as dt
import logging

from . import TadoLogger


def main():
    '''Launch the TadoLogger.'''
    parser = argparse.ArgumentParser()
    parser.add_argument('-o', '--outdir', help='path to output directory')
    parser.add_argument('--update-period', type=int,
                        help='time in seconds between updates',
                        default=30
                        )
    parser.add_argument('--last-day',
                        help='last day of week as 3 letter string',
                        default='sun',
                        )
    parser.add_argument('--disable-multiproc', action='store_false',
                        help="use single process for web query and csv saving",
                        dest='multiprocessing',
                        )
    parser.add_argument('-d', '--debug',
                        help='log debugging statements to file',
                        action='store_const', dest='loglevel', const=logging.DEBUG,
                        default=logging.WARNING,
                        )
    parser.add_argument('-v', '--verbose',
                        help='log info statements to file',
                        action='store_const', dest='loglevel', const=logging.INFO,
                        )
    args = parser.parse_args()
    with TadoLogger(args.outdir,
                    args.update_period,
                    args.last_day,
                    args.multiprocessing,
                    args.loglevel,
                    ) as tl:
        tl.start()
    print(f'Logging ended at {dt.datetime.now()}')


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import pathlib
from setuptools import setup, find_packages


with open(pathlib.Path(__file__).parent / 'README.md', 'r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(name='pytadolog',
      version='0.1.1',
      description='CSV logger for PyTado from Josh Gibson',
      long_description=long_description,
      long_description_content_type='text/markdown',
      keywords=['tado', 'python', 'logging'],
      author='Josh Gibson',
      author_email='josh-gibson@outlook.com',
      url='https://github.com/jjgibson/PyTadoLog',
      python_requires='>=3.6',
      install_requires=[
        'keyring',
        'numpy',
        'openpyxl',
        'pandas',
      ],
      license='MIT',
      platforms=['any'],
      packages=find_packages(),
      classifiers=[
        'Development Status :: 3 - Alpha',
        'Topic :: Home Automation',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6'
      ],
      entry_points ={ 
        'console_scripts': [ 
          'tadolog = pytadolog.__main__:main'
        ]
      },
      scripts=['scripts/csv2excel'],
      zip_safe=False,
)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

'''Contains utilities for the TadoLogger package.'''

import datetime as dt
import logging
import logging.handlers
import pathlib

import pandas as pd


def setupoutdir(outdir=None):
    '''Create directory structure for csvs and logs.

    Args:
        outdir (path like, optional): Path to top level directory. Defaults to None.
    
    If outdir is None <home>/Documents/TadoLogs/ is used.
    '''
    if outdir is None:
        outdir = pathlib.Path.home() / 'Documents' / 'TadoLogs'
    else:
        outdir = pathlib.Path(outdir)
    outdir.mkdir(exist_ok=True)
    return outdir


def setuplogger(logger, queue, level=logging.WARNING):
    '''Sets up the TadoLogger logging object.

    Args:
        logger (logging.logger): Python logging object to log to.
        queue (multiprocessing.Queue): Queue to log events to.
        level (logging.level, optional): Global level to log to. Defaults to
            logging.WARNING.
    '''
    logger.setLevel(level)
    logger.addHandler(logging.handlers.QueueHandler(queue))


def getnextday(day='sun'):
    '''Gets date of next occurence of day.

    Args:
        day (str, optional): Three letter weekday to get. Defaults to 'sun'.

    Returns:
        datetime.date: Date of next occurence of day.
    '''
    days = ('mon', 'tue', 'wed', 'thu', 'fri', 'sat', 'sun')
    today = dt.date.today()
    offset = (days.index(day.lower()) - today.weekday()) % 7
    # Add 30 seconds so next Sun is returned if it is between Sat 23:59:30 and
    # Sat 23:59:59
    return today + dt.timedelta(days=offset, seconds=30)


def ceildt(inp_dt, delta=30):
    '''Round the input up to the nearest delta in seconds.

    Args:
        inp_dt (datetime.datetime: Input datetime object to round.
        delta (int, optional): Seconds to round the input to. Defaults to 30.

    Returns:
        datetime.datetime: Rounded up datetime.

    Taken from https://stackoverflow.com/a/32724959
    '''
    return inp_dt + (dt.datetime.min - inp_dt) % delta


def nexttick(delta):
    '''Get the next time rounded to delta.

    Returns:
        datetime.datetime: Next time rounded to delta.
    '''
    return ceildt(dt.datetime.now(), dt.timedelta(seconds=delta))


def csv2excel(inpath, outpath=None, dropna=True):
    '''Convert TaDo CSV to Excel file.

    Args:
        inpath (path like): Path to csv file to be converted.
        outpath (path like, optional): Path to output Excel file.
            Defaults to None which uses input file path.
        dropna (bool, optional): Drop NaN rows on save. Defaults to True.
    '''
    df = pd.read_csv(inpath, header=[0, 1], index_col=0, parse_dates=True)
    if outpath is None:
        inpath = pathlib.Path(inpath)
        outpath = inpath.parent / inpath.with_suffix('.xlsx')
    if dropna:
        df.dropna(how='all').to_excel(outpath, freeze_panes=(2, 1))
    else:
        df.to_excel(outpath, freeze_panes=(2, 1))


def loglistener(outdir, queue, stopev, loglevel=logging.INFO):
    '''Collate logs from multiple processes.

    Args:
        outdir (path like): Top level directory in which to store log files.
        queue (Queue): Queue containing logging messages.
        stopev (Event): Event indicating that the process should be closed.
        loglevel (logging.level, optional): Level to log at. Defaults to 
            logging.INFO.
    
    Log files are stored in outdir/logs.
    '''
    outdir = outdir / 'logs'
    outdir.mkdir(exist_ok=True)
    root = logging.getLogger()
    h = logging.handlers.RotatingFileHandler(outdir / 'PyTadoLog.log', 'a', 500e3, 10)
    f = logging.Formatter(
        '[%(asctime)s] [%(processName)-15.15s] [%(name)-10.10s] '
        '[%(levelname)-8s] [%(funcName)-12.12s] %(message)s',
        '%Y-%m-%d %H:%M:%S')
    h.setFormatter(f)
    root.addHandler(h)
    root.setLevel(loglevel)

    while not stopev.is_set():
        try:
            record = queue.get()
            logger = logging.getLogger(record.name)
            logger.handle(record)
        except:
            root.exception('Exception encountered while handling logs')
